# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import sys

import numpy

import awkward as ak


def convert_to_array(layout, args, kwargs):
    out = ak.operations.convert.to_numpy(layout, allow_missing=False)
    if args == () and kwargs == {}:
        return out
    else:
        return numpy.array(out, *args, **kwargs)


implemented = {}


def _to_rectilinear(arg):
    if isinstance(
        arg,
        (
            ak.Array,
            ak.Record,
            ak.ArrayBuilder,
            ak.layout.Content,
            ak.layout.Record,
            ak.layout.ArrayBuilder,
            ak.layout.LayoutBuilder32,
            ak.layout.LayoutBuilder64,
        ),
    ):
        nplike = ak.nplike.of(arg)
        return nplike.to_rectilinear(arg)
    else:
        return arg


def array_function(func, types, args, kwargs):
    function = implemented.get(func)
    if function is None:
        args = tuple(_to_rectilinear(x) for x in args)
        kwargs = dict((k, _to_rectilinear(v)) for k, v in kwargs.items())
        out = func(*args, **kwargs)
        nplike = ak.nplike.of(out)
        if isinstance(out, nplike.ndarray) and len(out.shape) != 0:
            return ak.Array(out)
        else:
            return out
    else:
        return function(*args, **kwargs)


def implements(numpy_function):
    def decorator(function):
        implemented[getattr(numpy, numpy_function)] = function
        return function

    return decorator


def array_ufunc(ufunc, method, inputs, kwargs):
    if method != "__call__" or len(inputs) == 0 or "out" in kwargs:
        return NotImplemented

    behavior = ak._util.behaviorof(*inputs)

    nextinputs = []
    for x in inputs:
        cast_fcn = ak._util.custom_cast(x, behavior)
        if cast_fcn is not None:
            x = cast_fcn(x)
        nextinputs.append(
            ak.operations.convert.to_layout(x, allow_record=True, allow_other=True)
        )
    inputs = nextinputs

    def adjust(custom, inputs, kwargs):
        args = [
            ak._util.wrap(x, behavior)
            if isinstance(x, (ak.layout.Content, ak.layout.Record))
            else x
            for x in inputs
        ]
        out = custom(*args, **kwargs)
        if not isinstance(out, tuple):
            out = (out,)

        return tuple(
            x.layout if isinstance(x, (ak.highlevel.Array, ak.highlevel.Record)) else x
            for x in out
        )

    def adjust_apply_ufunc(apply_ufunc, ufunc, method, inputs, kwargs):
        nextinputs = [
            ak._util.wrap(x, behavior)
            if isinstance(x, (ak.layout.Content, ak.layout.Record))
            else x
            for x in inputs
        ]

        out = apply_ufunc(ufunc, method, nextinputs, kwargs)

        if out is NotImplemented:
            return None
        else:
            if not isinstance(out, tuple):
                out = (out,)
            out = tuple(
                x.layout
                if isinstance(x, (ak.highlevel.Array, ak.highlevel.Record))
                else x
                for x in out
            )
            return lambda: out

    def is_fully_regular(layout):
        if (
            isinstance(layout, ak.layout.RegularArray)
            and layout.parameter("__record__") is None
            and layout.parameter("__array__") is None
        ):
            if isinstance(layout.content, ak.layout.NumpyArray):
                return True
            elif isinstance(layout.content, ak.layout.RegularArray):
                return is_fully_regular(layout.content)
            else:
                return False
        else:
            return False

    def deregulate(layout):
        if not is_fully_regular(layout):
            return layout
        else:
            shape = [len(layout)]
            node = layout
            while isinstance(node, ak.layout.RegularArray):
                shape.append(node.size)
                node = node.content
            if node.format.upper().startswith("M"):
                nparray = ak.nplike.of(node).asarray(node.view_int64).view(node.format)
                nparray = nparray.reshape(tuple(shape) + nparray.shape[1:])
                return ak.layout.NumpyArray(
                    nparray,
                    node.identities,
                    node.parameters,
                )
            else:
                nparray = ak.nplike.of(node).asarray(node)
                nparray = nparray.reshape(tuple(shape) + nparray.shape[1:])
                return ak.layout.NumpyArray(
                    nparray,
                    node.identities,
                    node.parameters,
                )

    def getfunction(inputs):
        signature = [ufunc]
        for x in inputs:
            if isinstance(x, ak.layout.Content):
                record = x.parameter("__record__")
                array = x.parameter("__array__")
                if record is not None:
                    signature.append(record)
                elif array is not None:
                    signature.append(array)
                elif isinstance(x, ak.layout.NumpyArray):
                    if x.format.upper().startswith("M"):
                        signature.append(
                            ak.nplike.of(x)
                            .asarray(x.view_int64)
                            .view(x.format)
                            .dtype.type
                        )
                    else:
                        signature.append(ak.nplike.of(x).asarray(x).dtype.type)
                else:
                    signature.append(None)
            else:
                signature.append(type(x))

        custom = ak._util.overload(behavior, signature)
        if custom is not None:
            return lambda: adjust(custom, inputs, kwargs)

        if ufunc is numpy.matmul:
            custom_matmul = getfunction_matmul(inputs)
            if custom_matmul is not None:
                return custom_matmul

        inputs = [deregulate(x) for x in inputs]

        if all(
            (
                isinstance(x, ak.layout.NumpyArray)
                and not (x.format.upper().startswith("M"))
            )
            or not isinstance(x, (ak.layout.Content, ak.partition.PartitionedArray))
            for x in inputs
        ):
            nplike = ak.nplike.of(*inputs)
            result = getattr(ufunc, method)(
                *[nplike.asarray(x) for x in inputs], **kwargs
            )
            return lambda: (ak.operations.convert.from_numpy(result, highlevel=False),)
        elif all(
            isinstance(x, ak.layout.NumpyArray) and (x.format.upper().startswith("M"))
            for x in inputs
        ):
            nplike = ak.nplike.of(*inputs)
            result = getattr(ufunc, method)(
                *[nplike.asarray(x.view_int64).view(x.format) for x in inputs], **kwargs
            )
            return lambda: (ak.operations.convert.from_numpy(result, highlevel=False),)

        for x in inputs:
            if isinstance(x, ak.layout.Content):
                chained_behavior = ak._util.Behavior(ak.behavior, behavior)
                apply_ufunc = chained_behavior[numpy.ufunc, x.parameter("__array__")]
                if apply_ufunc is not None:
                    out = adjust_apply_ufunc(apply_ufunc, ufunc, method, inputs, kwargs)
                    if out is not None:
                        return out
                apply_ufunc = chained_behavior[numpy.ufunc, x.parameter("__record__")]
                if apply_ufunc is not None:
                    out = adjust_apply_ufunc(apply_ufunc, ufunc, method, inputs, kwargs)
                    if out is not None:
                        return out

        if all(
            x.parameter("__array__") is not None
            or x.parameter("__record__") is not None
            for x in inputs
            if isinstance(x, ak.layout.Content)
        ):
            custom_types = []
            for x in inputs:
                if isinstance(x, ak.layout.Content):
                    if x.parameter("__array__") is not None:
                        custom_types.append(x.parameter("__array__"))
                    elif x.parameter("__record__") is not None:
                        custom_types.append(x.parameter("__record__"))
                    else:
                        custom_types.append(type(x).__name__)
                else:
                    custom_types.append(type(x).__name__)
            raise ValueError(
                "no overloads for custom types: {0}({1})".format(
                    ufunc.__name__,
                    ", ".join(custom_types),
                )
                + ak._util.exception_suffix(__file__)
            )

        return None

    out = ak._util.broadcast_and_apply(
        inputs, getfunction, behavior, allow_records=False, pass_depth=False
    )
    assert isinstance(out, tuple) and len(out) == 1
    return ak._util.wrap(out[0], behavior)


def matmul_for_numba(lefts, rights, dtype):
    total_outer = 0
    total_inner = 0
    total_content = 0

    for A, B in zip(lefts, rights):
        first = -1
        for Ai in A:
            if first == -1:
                first = len(Ai)
            elif first != len(Ai):
                raise ValueError(
                    "one of the left matrices in np.matmul is not rectangular"
                )
        if first == -1:
            first = 0
        rowsA = len(A)
        colsA = first

        first = -1
        for Bi in B:
            if first == -1:
                first = len(Bi)
            elif first != len(Bi):
                raise ValueError(
                    "one of the right matrices in np.matmul is not rectangular"
                )
        if first == -1:
            first = 0
        rowsB = len(B)
        colsB = first

        if colsA != rowsB:
            raise ValueError(
                u"one of the pairs of matrices in np.matmul do not match shape: "
                u"(n \u00d7 k) @ (k \u00d7 m)"
            )

        total_outer += 1
        total_inner += rowsA
        total_content += rowsA * colsB

    outer = numpy.empty(total_outer + 1, numpy.int64)
    inner = numpy.empty(total_inner + 1, numpy.int64)
    content = numpy.zeros(total_content, dtype)

    outer[0] = 0
    inner[0] = 0
    outer_i = 1
    inner_i = 1
    content_i = 0
    for A, B in zip(lefts, rights):
        rows = len(A)
        cols = 0
        if len(B) > 0:
            cols = len(B[0])
        mids = 0
        if len(A) > 0:
            mids = len(A[0])

        for i in range(rows):
            for j in range(cols):
                for v in range(mids):
                    pos = content_i + i * cols + j
                    content[pos] += A[i][v] * B[v][j]

        outer[outer_i] = outer[outer_i - 1] + rows
        outer_i += 1
        for _ in range(rows):
            inner[inner_i] = inner[inner_i - 1] + cols
            inner_i += 1
        content_i += rows * cols

    return outer, inner, content


matmul_for_numba.numbafied = None


def getfunction_matmul(inputs):
    inputs = [
        ak._util.recursively_apply(
            x, (lambda _: _), pass_depth=False, numpy_to_regular=True
        )
        if isinstance(x, (ak.layout.Content, ak.layout.Record))
        else x
        for x in inputs
    ]

    if len(inputs) == 2 and all(
        isinstance(x, ak._util.listtypes)
        and isinstance(x.content, ak._util.listtypes)
        and isinstance(x.content.content, ak.layout.NumpyArray)
        for x in inputs
    ):
        ak._connect._numba.register_and_check()
        import numba

        if matmul_for_numba.numbafied is None:
            matmul_for_numba.numbafied = numba.njit(matmul_for_numba)

        lefts = ak.highlevel.Array(inputs[0])
        rights = ak.highlevel.Array(inputs[1])
        dtype = numpy.asarray(lefts[0:0, 0:0, 0:0] + rights[0:0, 0:0, 0:0]).dtype

        outer, inner, content = matmul_for_numba.numbafied(lefts, rights, dtype)

        return lambda: (
            ak.layout.ListOffsetArray64(
                ak.layout.Index64(outer),
                ak.layout.ListOffsetArray64(
                    ak.layout.Index64(inner),
                    ak.layout.NumpyArray(content),
                ),
            ),
        )

    else:
        return None


try:
    NDArrayOperatorsMixin = numpy.lib.mixins.NDArrayOperatorsMixin

except AttributeError:
    from numpy.core import umath as um

    def _disables_array_ufunc(obj):
        try:
            return obj.__array_ufunc__ is None
        except AttributeError:
            return False

    def _binary_method(ufunc, name):
        def func(self, other):
            if _disables_array_ufunc(other):
                return NotImplemented
            return ufunc(self, other)

        func.__name__ = "__{}__".format(name)
        return func

    def _reflected_binary_method(ufunc, name):
        def func(self, other):
            if _disables_array_ufunc(other):
                return NotImplemented
            return ufunc(other, self)

        func.__name__ = "__r{}__".format(name)
        return func

    def _inplace_binary_method(ufunc, name):
        def func(self, other):
            return ufunc(self, other, out=(self,))

        func.__name__ = "__i{}__".format(name)
        return func

    def _numeric_methods(ufunc, name):
        return (
            _binary_method(ufunc, name),
            _reflected_binary_method(ufunc, name),
            _inplace_binary_method(ufunc, name),
        )

    def _unary_method(ufunc, name):
        def func(self):
            return ufunc(self)

        func.__name__ = "__{}__".format(name)
        return func

    class NDArrayOperatorsMixin(object):
        __lt__ = _binary_method(um.less, "lt")
        __le__ = _binary_method(um.less_equal, "le")
        __eq__ = _binary_method(um.equal, "eq")
        __ne__ = _binary_method(um.not_equal, "ne")
        __gt__ = _binary_method(um.greater, "gt")
        __ge__ = _binary_method(um.greater_equal, "ge")

        __add__, __radd__, __iadd__ = _numeric_methods(um.add, "add")
        __sub__, __rsub__, __isub__ = _numeric_methods(um.subtract, "sub")
        __mul__, __rmul__, __imul__ = _numeric_methods(um.multiply, "mul")
        __matmul__, __rmatmul__, __imatmul__ = _numeric_methods(um.matmul, "matmul")
        if sys.version_info.major < 3:
            __div__, __rdiv__, __idiv__ = _numeric_methods(um.divide, "div")
        __truediv__, __rtruediv__, __itruediv__ = _numeric_methods(
            um.true_divide, "truediv"
        )
        __floordiv__, __rfloordiv__, __ifloordiv__ = _numeric_methods(
            um.floor_divide, "floordiv"
        )
        __mod__, __rmod__, __imod__ = _numeric_methods(um.remainder, "mod")
        if hasattr(um, "divmod"):
            __divmod__ = _binary_method(um.divmod, "divmod")
            __rdivmod__ = _reflected_binary_method(um.divmod, "divmod")
        __pow__, __rpow__, __ipow__ = _numeric_methods(um.power, "pow")
        __lshift__, __rlshift__, __ilshift__ = _numeric_methods(um.left_shift, "lshift")
        __rshift__, __rrshift__, __irshift__ = _numeric_methods(
            um.right_shift, "rshift"
        )
        __and__, __rand__, __iand__ = _numeric_methods(um.bitwise_and, "and")
        __xor__, __rxor__, __ixor__ = _numeric_methods(um.bitwise_xor, "xor")
        __or__, __ror__, __ior__ = _numeric_methods(um.bitwise_or, "or")

        __neg__ = _unary_method(um.negative, "neg")
        if hasattr(um, "positive"):
            __pos__ = _unary_method(um.positive, "pos")
        __abs__ = _unary_method(um.absolute, "abs")
        __invert__ = _unary_method(um.invert, "invert")



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401

jax = pytest.importorskip("jax")
jax.config.update("jax_platform_name", "gpu")
jax.config.update("jax_enable_x64", True)


def test_from_jax():
    jax_array_1d = jax.numpy.arange(10)
    jax_array_2d = jax.numpy.array([[1.1, 2.2], [3.3, 4.4], [5.5, 6.6], [7.7, 8.8]])

    ak_jax_array_1d = ak.from_jax(jax_array_1d)
    ak_jax_array_2d = ak.from_jax(jax_array_2d)

    for i in range(10):
        assert ak_jax_array_1d[i] == jax_array_1d[i]

    for i in range(4):
        for j in range(2):
            assert ak_jax_array_2d[i][j] == jax_array_2d[i][j]


def test_from_jax_tolist():
    jax_array_1d = jax.numpy.array([9, 8, 7, 6, 5, 4, 3, 2, 1, 0])

    ak_jax_array_1d = ak.from_jax(jax_array_1d)

    assert ak.to_list(ak_jax_array_1d.layout) == [9, 8, 7, 6, 5, 4, 3, 2, 1, 0]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_issue434():
    a = ak.Array([[0.0, 1.1, 2.2], [3.3, 4.4], [5.5]])
    b = ak.Array([[9.9, 8.8, 7.7], [6.6, 5.5], [4.4]])
    assert ak.to_list(b[ak.argmin(a, axis=1, keepdims=True)]) == [[9.9], [6.6], [4.4]]
    assert ak.to_list(b[ak.argmax(a, axis=1, keepdims=True)]) == [[7.7], [5.5], [4.4]]


def test_nokeepdims():
    nparray = np.arange(2 * 3 * 5, dtype=np.int64).reshape(2, 3, 5)
    content = ak.layout.NumpyArray(np.arange(2 * 3 * 5, dtype=np.int64))
    regular = ak.layout.RegularArray(content, 5, zeros_length=0)
    listoffset = regular.toListOffsetArray64(False)
    regular_regular = ak.layout.RegularArray(regular, 3, zeros_length=0)
    listoffset_regular = regular_regular.toListOffsetArray64(False)
    regular_listoffset = ak.layout.RegularArray(listoffset, 3, zeros_length=0)
    listoffset_listoffset = regular_listoffset.toListOffsetArray64(False)

    assert ak.to_list(regular_regular) == [
        [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9], [10, 11, 12, 13, 14]],
        [[15, 16, 17, 18, 19], [20, 21, 22, 23, 24], [25, 26, 27, 28, 29]],
    ]
    assert ak.to_list(listoffset_regular) == [
        [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9], [10, 11, 12, 13, 14]],
        [[15, 16, 17, 18, 19], [20, 21, 22, 23, 24], [25, 26, 27, 28, 29]],
    ]
    assert ak.to_list(regular_listoffset) == [
        [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9], [10, 11, 12, 13, 14]],
        [[15, 16, 17, 18, 19], [20, 21, 22, 23, 24], [25, 26, 27, 28, 29]],
    ]
    assert ak.to_list(listoffset_listoffset) == [
        [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9], [10, 11, 12, 13, 14]],
        [[15, 16, 17, 18, 19], [20, 21, 22, 23, 24], [25, 26, 27, 28, 29]],
    ]

    assert str(ak.type(ak.Array(listoffset_listoffset))) == "2 * var * var * int64"
    axis1 = ak.sum(listoffset_listoffset, axis=-1)
    axis2 = ak.sum(listoffset_listoffset, axis=-2)
    axis3 = ak.sum(listoffset_listoffset, axis=-3)
    assert ak.to_list(axis1) == np.sum(nparray, axis=-1).tolist()
    assert ak.to_list(axis2) == np.sum(nparray, axis=-2).tolist()
    assert ak.to_list(axis3) == np.sum(nparray, axis=-3).tolist()
    assert str(ak.type(ak.Array(axis1))) == "2 * var * int64"
    assert str(ak.type(ak.Array(axis2))) == "2 * var * int64"
    assert str(ak.type(ak.Array(axis3))) == "3 * var * int64"

    assert str(ak.type(ak.Array(listoffset_regular))) == "2 * var * 5 * int64"
    axis1 = ak.sum(listoffset_regular, axis=-1)
    axis2 = ak.sum(listoffset_regular, axis=-2)
    axis3 = ak.sum(listoffset_regular, axis=-3)
    assert ak.to_list(axis1) == np.sum(nparray, axis=-1).tolist()
    assert ak.to_list(axis2) == np.sum(nparray, axis=-2).tolist()
    assert ak.to_list(axis3) == np.sum(nparray, axis=-3).tolist()
    assert str(ak.type(ak.Array(axis1))) == "2 * var * int64"
    assert str(ak.type(ak.Array(axis2))) == "2 * 5 * int64"
    assert str(ak.type(ak.Array(axis3))) == "3 * 5 * int64"

    assert str(ak.type(ak.Array(regular_listoffset))) == "2 * 3 * var * int64"
    axis1 = ak.sum(regular_listoffset, axis=-1)
    axis2 = ak.sum(regular_listoffset, axis=-2)
    axis3 = ak.sum(regular_listoffset, axis=-3)
    assert ak.to_list(axis1) == np.sum(nparray, axis=-1).tolist()
    assert ak.to_list(axis2) == np.sum(nparray, axis=-2).tolist()
    assert ak.to_list(axis3) == np.sum(nparray, axis=-3).tolist()
    assert str(ak.type(ak.Array(axis1))) == "2 * 3 * int64"
    assert str(ak.type(ak.Array(axis2))) == "2 * var * int64"
    assert str(ak.type(ak.Array(axis3))) == "3 * var * int64"

    assert str(ak.type(ak.Array(regular_regular))) == "2 * 3 * 5 * int64"
    axis1 = ak.sum(regular_regular, axis=-1)
    axis2 = ak.sum(regular_regular, axis=-2)
    axis3 = ak.sum(regular_regular, axis=-3)
    assert ak.to_list(axis1) == np.sum(nparray, axis=-1).tolist()
    assert ak.to_list(axis2) == np.sum(nparray, axis=-2).tolist()
    assert ak.to_list(axis3) == np.sum(nparray, axis=-3).tolist()
    assert str(ak.type(ak.Array(axis1))) == "2 * 3 * int64"
    assert str(ak.type(ak.Array(axis2))) == "2 * 5 * int64"
    assert str(ak.type(ak.Array(axis3))) == "3 * 5 * int64"


def test_keepdims():
    nparray = np.arange(2 * 3 * 5, dtype=np.int64).reshape(2, 3, 5)
    content = ak.layout.NumpyArray(np.arange(2 * 3 * 5, dtype=np.int64))
    regular = ak.layout.RegularArray(content, 5, zeros_length=0)
    listoffset = regular.toListOffsetArray64(False)
    regular_regular = ak.layout.RegularArray(regular, 3, zeros_length=0)
    listoffset_regular = regular_regular.toListOffsetArray64(False)
    regular_listoffset = ak.layout.RegularArray(listoffset, 3, zeros_length=0)
    listoffset_listoffset = regular_listoffset.toListOffsetArray64(False)

    assert str(ak.type(ak.Array(listoffset_listoffset))) == "2 * var * var * int64"
    axis1 = ak.sum(listoffset_listoffset, axis=-1, keepdims=True)
    axis2 = ak.sum(listoffset_listoffset, axis=-2, keepdims=True)
    axis3 = ak.sum(listoffset_listoffset, axis=-3, keepdims=True)
    assert ak.to_list(axis1) == np.sum(nparray, axis=-1, keepdims=True).tolist()
    assert ak.to_list(axis2) == np.sum(nparray, axis=-2, keepdims=True).tolist()
    assert ak.to_list(axis3) == np.sum(nparray, axis=-3, keepdims=True).tolist()
    assert str(ak.type(ak.Array(axis1))) == "2 * var * var * int64"
    assert str(ak.type(ak.Array(axis2))) == "2 * var * var * int64"
    assert str(ak.type(ak.Array(axis3))) == "1 * var * var * int64"

    assert str(ak.type(ak.Array(listoffset_regular))) == "2 * var * 5 * int64"
    axis1 = ak.sum(listoffset_regular, axis=-1, keepdims=True)
    axis2 = ak.sum(listoffset_regular, axis=-2, keepdims=True)
    axis3 = ak.sum(listoffset_regular, axis=-3, keepdims=True)
    assert ak.to_list(axis1) == np.sum(nparray, axis=-1, keepdims=True).tolist()
    assert ak.to_list(axis2) == np.sum(nparray, axis=-2, keepdims=True).tolist()
    assert ak.to_list(axis3) == np.sum(nparray, axis=-3, keepdims=True).tolist()
    assert str(ak.type(ak.Array(axis1))) == "2 * var * 1 * int64"
    assert str(ak.type(ak.Array(axis2))) == "2 * var * 5 * int64"
    assert str(ak.type(ak.Array(axis3))) == "1 * var * 5 * int64"

    assert str(ak.type(ak.Array(regular_listoffset))) == "2 * 3 * var * int64"
    axis1 = ak.sum(regular_listoffset, axis=-1, keepdims=True)
    axis2 = ak.sum(regular_listoffset, axis=-2, keepdims=True)
    axis3 = ak.sum(regular_listoffset, axis=-3, keepdims=True)
    assert ak.to_list(axis1) == np.sum(nparray, axis=-1, keepdims=True).tolist()
    assert ak.to_list(axis2) == np.sum(nparray, axis=-2, keepdims=True).tolist()
    assert ak.to_list(axis3) == np.sum(nparray, axis=-3, keepdims=True).tolist()
    assert str(ak.type(ak.Array(axis1))) == "2 * 3 * var * int64"
    assert str(ak.type(ak.Array(axis2))) == "2 * 1 * var * int64"
    assert str(ak.type(ak.Array(axis3))) == "1 * 3 * var * int64"

    assert str(ak.type(ak.Array(regular_regular))) == "2 * 3 * 5 * int64"
    axis1 = ak.sum(regular_regular, axis=-1, keepdims=True)
    axis2 = ak.sum(regular_regular, axis=-2, keepdims=True)
    axis3 = ak.sum(regular_regular, axis=-3, keepdims=True)
    assert ak.to_list(axis1) == np.sum(nparray, axis=-1, keepdims=True).tolist()
    assert ak.to_list(axis2) == np.sum(nparray, axis=-2, keepdims=True).tolist()
    assert ak.to_list(axis3) == np.sum(nparray, axis=-3, keepdims=True).tolist()
    assert str(ak.type(ak.Array(axis1))) == "2 * 3 * 1 * int64"
    assert str(ak.type(ak.Array(axis2))) == "2 * 1 * 5 * int64"
    assert str(ak.type(ak.Array(axis3))) == "1 * 3 * 5 * int64"


def test_nokeepdims_none1():
    content = ak.Array(
        [
            0,
            1,
            2,
            None,
            4,
            5,
            None,
            None,
            8,
            9,
            10,
            11,
            12,
            None,
            14,
            15,
            16,
            17,
            18,
            None,
            None,
            None,
            None,
            None,
            None,
            25,
            26,
            27,
            28,
            29,
        ]
    ).layout
    regular = ak.layout.RegularArray(content, 5, zeros_length=0)
    listoffset = regular.toListOffsetArray64(False)
    regular_regular = ak.layout.RegularArray(regular, 3, zeros_length=0)
    listoffset_regular = regular_regular.toListOffsetArray64(False)
    regular_listoffset = ak.layout.RegularArray(listoffset, 3, zeros_length=0)
    listoffset_listoffset = regular_listoffset.toListOffsetArray64(False)

    assert str(ak.type(ak.Array(listoffset_listoffset))) == "2 * var * var * ?int64"
    axis1 = ak.sum(listoffset_listoffset, axis=-1)
    axis2 = ak.sum(listoffset_listoffset, axis=-2)
    axis3 = ak.sum(listoffset_listoffset, axis=-3)
    assert str(ak.type(ak.Array(axis1))) == "2 * var * int64"
    assert str(ak.type(ak.Array(axis2))) == "2 * var * int64"
    assert str(ak.type(ak.Array(axis3))) == "3 * var * int64"

    assert str(ak.type(ak.Array(listoffset_regular))) == "2 * var * 5 * ?int64"
    axis1 = ak.sum(listoffset_regular, axis=-1)
    axis2 = ak.sum(listoffset_regular, axis=-2)
    axis3 = ak.sum(listoffset_regular, axis=-3)
    assert str(ak.type(ak.Array(axis1))) == "2 * var * int64"
    assert str(ak.type(ak.Array(axis2))) == "2 * var * int64"
    assert str(ak.type(ak.Array(axis3))) == "3 * var * int64"

    assert str(ak.type(ak.Array(regular_listoffset))) == "2 * 3 * var * ?int64"
    axis1 = ak.sum(regular_listoffset, axis=-1)
    axis2 = ak.sum(regular_listoffset, axis=-2)
    axis3 = ak.sum(regular_listoffset, axis=-3)
    assert str(ak.type(ak.Array(axis1))) == "2 * 3 * int64"
    assert str(ak.type(ak.Array(axis2))) == "2 * var * int64"
    assert str(ak.type(ak.Array(axis3))) == "3 * var * int64"

    assert str(ak.type(ak.Array(regular_regular))) == "2 * 3 * 5 * ?int64"
    axis1 = ak.sum(regular_regular, axis=-1)
    axis2 = ak.sum(regular_regular, axis=-2)
    axis3 = ak.sum(regular_regular, axis=-3)
    assert str(ak.type(ak.Array(axis1))) == "2 * 3 * int64"
    assert str(ak.type(ak.Array(axis2))) == "2 * var * int64"
    assert str(ak.type(ak.Array(axis3))) == "3 * var * int64"


def test_keepdims_none1():
    content = ak.Array(
        [
            0,
            1,
            2,
            None,
            4,
            5,
            None,
            None,
            8,
            9,
            10,
            11,
            12,
            None,
            14,
            15,
            16,
            17,
            18,
            None,
            None,
            None,
            None,
            None,
            None,
            25,
            26,
            27,
            28,
            29,
        ]
    ).layout
    regular = ak.layout.RegularArray(content, 5, zeros_length=0)
    listoffset = regular.toListOffsetArray64(False)
    regular_regular = ak.layout.RegularArray(regular, 3, zeros_length=0)
    listoffset_regular = regular_regular.toListOffsetArray64(False)
    regular_listoffset = ak.layout.RegularArray(listoffset, 3, zeros_length=0)
    listoffset_listoffset = regular_listoffset.toListOffsetArray64(False)

    assert str(ak.type(ak.Array(listoffset_listoffset))) == "2 * var * var * ?int64"
    axis1 = ak.sum(listoffset_listoffset, axis=-1, keepdims=True)
    axis2 = ak.sum(listoffset_listoffset, axis=-2, keepdims=True)
    axis3 = ak.sum(listoffset_listoffset, axis=-3, keepdims=True)
    assert str(ak.type(ak.Array(axis1))) == "2 * var * var * int64"
    assert str(ak.type(ak.Array(axis2))) == "2 * var * var * int64"
    assert str(ak.type(ak.Array(axis3))) == "1 * var * var * int64"

    assert str(ak.type(ak.Array(listoffset_regular))) == "2 * var * 5 * ?int64"
    axis1 = ak.sum(listoffset_regular, axis=-1, keepdims=True)
    axis2 = ak.sum(listoffset_regular, axis=-2, keepdims=True)
    axis3 = ak.sum(listoffset_regular, axis=-3, keepdims=True)
    assert str(ak.type(ak.Array(axis1))) == "2 * var * var * int64"
    assert str(ak.type(ak.Array(axis2))) == "2 * var * var * int64"
    assert str(ak.type(ak.Array(axis3))) == "1 * var * var * int64"

    assert str(ak.type(ak.Array(regular_listoffset))) == "2 * 3 * var * ?int64"
    axis1 = ak.sum(regular_listoffset, axis=-1, keepdims=True)
    axis2 = ak.sum(regular_listoffset, axis=-2, keepdims=True)
    axis3 = ak.sum(regular_listoffset, axis=-3, keepdims=True)
    assert str(ak.type(ak.Array(axis1))) == "2 * 3 * var * int64"
    assert str(ak.type(ak.Array(axis2))) == "2 * 1 * var * int64"
    assert str(ak.type(ak.Array(axis3))) == "1 * 3 * var * int64"

    assert str(ak.type(ak.Array(regular_regular))) == "2 * 3 * 5 * ?int64"
    axis1 = ak.sum(regular_regular, axis=-1, keepdims=True)
    axis2 = ak.sum(regular_regular, axis=-2, keepdims=True)
    axis3 = ak.sum(regular_regular, axis=-3, keepdims=True)
    assert str(ak.type(ak.Array(axis1))) == "2 * 3 * var * int64"
    assert str(ak.type(ak.Array(axis2))) == "2 * 1 * var * int64"
    assert str(ak.type(ak.Array(axis3))) == "1 * 3 * var * int64"


def test_nokeepdims_mask1():
    mask = ak.layout.Index8(
        np.array(
            [
                False,
                False,
                False,
                True,
                False,
                False,
                True,
                True,
                False,
                False,
                False,
                False,
                False,
                True,
                False,
                False,
                False,
                False,
                False,
                True,
                True,
                True,
                True,
                True,
                True,
                False,
                False,
                False,
                False,
                False,
            ]
        )
    )
    content = ak.layout.ByteMaskedArray(
        mask,
        ak.layout.NumpyArray(np.arange(2 * 3 * 5, dtype=np.int64)),
        valid_when=False,
    )
    regular = ak.layout.RegularArray(content, 5, zeros_length=0)
    listoffset = regular.toListOffsetArray64(False)
    regular_regular = ak.layout.RegularArray(regular, 3, zeros_length=0)
    listoffset_regular = regular_regular.toListOffsetArray64(False)
    regular_listoffset = ak.layout.RegularArray(listoffset, 3, zeros_length=0)
    listoffset_listoffset = regular_listoffset.toListOffsetArray64(False)

    assert str(ak.type(ak.Array(listoffset_listoffset))) == "2 * var * var * ?int64"
    axis1 = ak.sum(listoffset_listoffset, axis=-1)
    axis2 = ak.sum(listoffset_listoffset, axis=-2)
    axis3 = ak.sum(listoffset_listoffset, axis=-3)
    assert str(ak.type(ak.Array(axis1))) == "2 * var * int64"
    assert str(ak.type(ak.Array(axis2))) == "2 * var * int64"
    assert str(ak.type(ak.Array(axis3))) == "3 * var * int64"

    assert str(ak.type(ak.Array(listoffset_regular))) == "2 * var * 5 * ?int64"
    axis1 = ak.sum(listoffset_regular, axis=-1)
    axis2 = ak.sum(listoffset_regular, axis=-2)
    axis3 = ak.sum(listoffset_regular, axis=-3)
    assert str(ak.type(ak.Array(axis1))) == "2 * var * int64"
    assert str(ak.type(ak.Array(axis2))) == "2 * var * int64"
    assert str(ak.type(ak.Array(axis3))) == "3 * var * int64"

    assert str(ak.type(ak.Array(regular_listoffset))) == "2 * 3 * var * ?int64"
    axis1 = ak.sum(regular_listoffset, axis=-1)
    axis2 = ak.sum(regular_listoffset, axis=-2)
    axis3 = ak.sum(regular_listoffset, axis=-3)
    assert str(ak.type(ak.Array(axis1))) == "2 * 3 * int64"
    assert str(ak.type(ak.Array(axis2))) == "2 * var * int64"
    assert str(ak.type(ak.Array(axis3))) == "3 * var * int64"

    assert str(ak.type(ak.Array(regular_regular))) == "2 * 3 * 5 * ?int64"
    axis1 = ak.sum(regular_regular, axis=-1)
    axis2 = ak.sum(regular_regular, axis=-2)
    axis3 = ak.sum(regular_regular, axis=-3)
    assert str(ak.type(ak.Array(axis1))) == "2 * 3 * int64"
    assert str(ak.type(ak.Array(axis2))) == "2 * var * int64"
    assert str(ak.type(ak.Array(axis3))) == "3 * var * int64"


def test_keepdims_mask1():
    mask = ak.layout.Index8(
        np.array(
            [
                False,
                False,
                False,
                True,
                False,
                False,
                True,
                True,
                False,
                False,
                False,
                False,
                False,
                True,
                False,
                False,
                False,
                False,
                False,
                True,
                True,
                True,
                True,
                True,
                True,
                False,
                False,
                False,
                False,
                False,
            ]
        )
    )
    content = ak.layout.ByteMaskedArray(
        mask,
        ak.layout.NumpyArray(np.arange(2 * 3 * 5, dtype=np.int64)),
        valid_when=False,
    )
    regular = ak.layout.RegularArray(content, 5, zeros_length=0)
    listoffset = regular.toListOffsetArray64(False)
    regular_regular = ak.layout.RegularArray(regular, 3, zeros_length=0)
    listoffset_regular = regular_regular.toListOffsetArray64(False)
    regular_listoffset = ak.layout.RegularArray(listoffset, 3, zeros_length=0)
    listoffset_listoffset = regular_listoffset.toListOffsetArray64(False)

    assert str(ak.type(ak.Array(listoffset_listoffset))) == "2 * var * var * ?int64"
    axis1 = ak.sum(listoffset_listoffset, axis=-1, keepdims=True)
    axis2 = ak.sum(listoffset_listoffset, axis=-2, keepdims=True)
    axis3 = ak.sum(listoffset_listoffset, axis=-3, keepdims=True)
    assert str(ak.type(ak.Array(axis1))) == "2 * var * var * int64"
    assert str(ak.type(ak.Array(axis2))) == "2 * var * var * int64"
    assert str(ak.type(ak.Array(axis3))) == "1 * var * var * int64"

    assert str(ak.type(ak.Array(listoffset_regular))) == "2 * var * 5 * ?int64"
    axis1 = ak.sum(listoffset_regular, axis=-1, keepdims=True)
    axis2 = ak.sum(listoffset_regular, axis=-2, keepdims=True)
    axis3 = ak.sum(listoffset_regular, axis=-3, keepdims=True)
    assert str(ak.type(ak.Array(axis1))) == "2 * var * var * int64"
    assert str(ak.type(ak.Array(axis2))) == "2 * var * var * int64"
    assert str(ak.type(ak.Array(axis3))) == "1 * var * var * int64"

    assert str(ak.type(ak.Array(regular_listoffset))) == "2 * 3 * var * ?int64"
    axis1 = ak.sum(regular_listoffset, axis=-1, keepdims=True)
    axis2 = ak.sum(regular_listoffset, axis=-2, keepdims=True)
    axis3 = ak.sum(regular_listoffset, axis=-3, keepdims=True)
    assert str(ak.type(ak.Array(axis1))) == "2 * 3 * var * int64"
    assert str(ak.type(ak.Array(axis2))) == "2 * 1 * var * int64"
    assert str(ak.type(ak.Array(axis3))) == "1 * 3 * var * int64"

    assert str(ak.type(ak.Array(regular_regular))) == "2 * 3 * 5 * ?int64"
    axis1 = ak.sum(regular_regular, axis=-1, keepdims=True)
    axis2 = ak.sum(regular_regular, axis=-2, keepdims=True)
    axis3 = ak.sum(regular_regular, axis=-3, keepdims=True)
    assert str(ak.type(ak.Array(axis1))) == "2 * 3 * var * int64"
    assert str(ak.type(ak.Array(axis2))) == "2 * 1 * var * int64"
    assert str(ak.type(ak.Array(axis3))) == "1 * 3 * var * int64"


def test_nokeepdims_mask2():
    content = ak.layout.NumpyArray(np.arange(2 * 3 * 5, dtype=np.int64))
    regular = ak.layout.RegularArray(content, 5, zeros_length=0)
    listoffset = regular.toListOffsetArray64(False)
    mask = ak.layout.Index8(np.array([False, False, True, True, False, True]))
    regular_regular = ak.layout.RegularArray(
        ak.layout.ByteMaskedArray(mask, regular, valid_when=False), 3, zeros_length=0
    )
    listoffset_regular = regular_regular.toListOffsetArray64(False)
    regular_listoffset = ak.layout.RegularArray(
        ak.layout.ByteMaskedArray(mask, listoffset, valid_when=False), 3, zeros_length=0
    )
    listoffset_listoffset = regular_listoffset.toListOffsetArray64(False)

    assert (
        str(ak.type(ak.Array(listoffset_listoffset))) == "2 * var * option[var * int64]"
    )
    axis1 = ak.sum(listoffset_listoffset, axis=-1)
    axis2 = ak.sum(listoffset_listoffset, axis=-2)
    axis3 = ak.sum(listoffset_listoffset, axis=-3)
    assert str(ak.type(ak.Array(axis1))) == "2 * var * ?int64"
    assert str(ak.type(ak.Array(axis2))) == "2 * var * int64"
    assert str(ak.type(ak.Array(axis3))) == "3 * var * int64"

    assert str(ak.type(ak.Array(listoffset_regular))) == "2 * var * option[5 * int64]"
    axis1 = ak.sum(listoffset_regular, axis=-1)
    axis2 = ak.sum(listoffset_regular, axis=-2)
    axis3 = ak.sum(listoffset_regular, axis=-3)
    assert str(ak.type(ak.Array(axis1))) == "2 * var * ?int64"
    assert str(ak.type(ak.Array(axis2))) == "2 * var * int64"
    assert str(ak.type(ak.Array(axis3))) == "3 * var * int64"

    assert str(ak.type(ak.Array(regular_listoffset))) == "2 * 3 * option[var * int64]"
    axis1 = ak.sum(regular_listoffset, axis=-1)
    axis2 = ak.sum(regular_listoffset, axis=-2)
    axis3 = ak.sum(regular_listoffset, axis=-3)
    assert str(ak.type(ak.Array(axis1))) == "2 * var * ?int64"
    assert str(ak.type(ak.Array(axis2))) == "2 * var * int64"
    assert str(ak.type(ak.Array(axis3))) == "3 * var * int64"

    assert str(ak.type(ak.Array(regular_regular))) == "2 * 3 * option[5 * int64]"
    axis1 = ak.sum(regular_regular, axis=-1)
    axis2 = ak.sum(regular_regular, axis=-2)
    axis3 = ak.sum(regular_regular, axis=-3)
    assert str(ak.type(ak.Array(axis1))) == "2 * var * ?int64"
    assert str(ak.type(ak.Array(axis2))) == "2 * var * int64"
    assert str(ak.type(ak.Array(axis3))) == "3 * var * int64"


def test_keepdims_mask2():
    content = ak.layout.NumpyArray(np.arange(2 * 3 * 5, dtype=np.int64))
    regular = ak.layout.RegularArray(content, 5, zeros_length=0)
    listoffset = regular.toListOffsetArray64(False)
    mask = ak.layout.Index8(np.array([False, False, True, True, False, True]))
    regular_regular = ak.layout.RegularArray(
        ak.layout.ByteMaskedArray(mask, regular, valid_when=False), 3, zeros_length=0
    )
    listoffset_regular = regular_regular.toListOffsetArray64(False)
    regular_listoffset = ak.layout.RegularArray(
        ak.layout.ByteMaskedArray(mask, listoffset, valid_when=False), 3, zeros_length=0
    )
    listoffset_listoffset = regular_listoffset.toListOffsetArray64(False)

    assert (
        str(ak.type(ak.Array(listoffset_listoffset))) == "2 * var * option[var * int64]"
    )
    axis1 = ak.sum(listoffset_listoffset, axis=-1, keepdims=True)
    axis2 = ak.sum(listoffset_listoffset, axis=-2, keepdims=True)
    axis3 = ak.sum(listoffset_listoffset, axis=-3, keepdims=True)
    assert str(ak.type(ak.Array(axis1))) == "2 * var * option[var * int64]"
    assert str(ak.type(ak.Array(axis2))) == "2 * var * var * int64"
    assert str(ak.type(ak.Array(axis3))) == "1 * var * var * int64"

    assert str(ak.type(ak.Array(listoffset_regular))) == "2 * var * option[5 * int64]"
    axis1 = ak.sum(listoffset_regular, axis=-1, keepdims=True)
    axis2 = ak.sum(listoffset_regular, axis=-2, keepdims=True)
    axis3 = ak.sum(listoffset_regular, axis=-3, keepdims=True)
    assert str(ak.type(ak.Array(axis1))) == "2 * var * option[1 * int64]"
    assert str(ak.type(ak.Array(axis2))) == "2 * var * var * int64"
    assert str(ak.type(ak.Array(axis3))) == "1 * var * var * int64"

    assert str(ak.type(ak.Array(regular_listoffset))) == "2 * 3 * option[var * int64]"
    axis1 = ak.sum(regular_listoffset, axis=-1, keepdims=True)
    axis2 = ak.sum(regular_listoffset, axis=-2, keepdims=True)
    axis3 = ak.sum(regular_listoffset, axis=-3, keepdims=True)
    assert str(ak.type(ak.Array(axis1))) == "2 * var * option[var * int64]"
    assert str(ak.type(ak.Array(axis2))) == "2 * var * var * int64"
    assert str(ak.type(ak.Array(axis3))) == "1 * var * var * int64"

    assert str(ak.type(ak.Array(regular_regular))) == "2 * 3 * option[5 * int64]"
    axis1 = ak.sum(regular_regular, axis=-1, keepdims=True)
    axis2 = ak.sum(regular_regular, axis=-2, keepdims=True)
    axis3 = ak.sum(regular_regular, axis=-3, keepdims=True)
    assert str(ak.type(ak.Array(axis1))) == "2 * var * option[1 * int64]"
    assert str(ak.type(ak.Array(axis2))) == "2 * var * var * int64"
    assert str(ak.type(ak.Array(axis3))) == "1 * var * var * int64"


def test_nokeepdims_mask3():
    content = ak.layout.NumpyArray(np.arange(2 * 3 * 5, dtype=np.int64))
    regular = ak.layout.RegularArray(content, 5, zeros_length=0)
    listoffset = regular.toListOffsetArray64(False)
    regular_regular = ak.layout.RegularArray(regular, 3, zeros_length=0)
    listoffset_regular = regular_regular.toListOffsetArray64(False)
    regular_listoffset = ak.layout.RegularArray(listoffset, 3, zeros_length=0)
    listoffset_listoffset = regular_listoffset.toListOffsetArray64(False)
    mask = ak.layout.Index8(np.array([True, False]))
    regular_regular = ak.layout.ByteMaskedArray(mask, regular_regular, valid_when=False)
    listoffset_regular = ak.layout.ByteMaskedArray(
        mask, listoffset_regular, valid_when=False
    )
    regular_listoffset = ak.layout.ByteMaskedArray(
        mask, regular_listoffset, valid_when=False
    )
    listoffset_listoffset = ak.layout.ByteMaskedArray(
        mask, listoffset_listoffset, valid_when=False
    )

    assert (
        str(ak.type(ak.Array(listoffset_listoffset))) == "2 * option[var * var * int64]"
    )
    axis1 = ak.sum(listoffset_listoffset, axis=-1)
    axis2 = ak.sum(listoffset_listoffset, axis=-2)
    axis3 = ak.sum(listoffset_listoffset, axis=-3)
    assert str(ak.type(ak.Array(axis1))) == "2 * option[var * int64]"
    assert str(ak.type(ak.Array(axis2))) == "2 * option[var * int64]"
    assert str(ak.type(ak.Array(axis3))) == "3 * var * int64"

    assert str(ak.type(ak.Array(listoffset_regular))) == "2 * option[var * 5 * int64]"
    axis1 = ak.sum(listoffset_regular, axis=-1)
    axis2 = ak.sum(listoffset_regular, axis=-2)
    axis3 = ak.sum(listoffset_regular, axis=-3)
    assert str(ak.type(ak.Array(axis1))) == "2 * option[var * int64]"
    assert str(ak.type(ak.Array(axis2))) == "2 * option[5 * int64]"
    assert str(ak.type(ak.Array(axis3))) == "3 * 5 * int64"

    assert str(ak.type(ak.Array(regular_listoffset))) == "2 * option[3 * var * int64]"
    axis1 = ak.sum(regular_listoffset, axis=-1)
    axis2 = ak.sum(regular_listoffset, axis=-2)
    axis3 = ak.sum(regular_listoffset, axis=-3)
    assert str(ak.type(ak.Array(axis1))) == "2 * option[var * int64]"
    assert str(ak.type(ak.Array(axis2))) == "2 * option[var * int64]"
    assert str(ak.type(ak.Array(axis3))) == "3 * var * int64"

    assert str(ak.type(ak.Array(regular_regular))) == "2 * option[3 * 5 * int64]"
    axis1 = ak.sum(regular_regular, axis=-1)
    axis2 = ak.sum(regular_regular, axis=-2)
    axis3 = ak.sum(regular_regular, axis=-3)
    assert str(ak.type(ak.Array(axis1))) == "2 * option[var * int64]"
    assert str(ak.type(ak.Array(axis2))) == "2 * option[5 * int64]"
    assert str(ak.type(ak.Array(axis3))) == "3 * 5 * int64"


def test_keepdims_mask3():
    content = ak.layout.NumpyArray(np.arange(2 * 3 * 5, dtype=np.int64))
    regular = ak.layout.RegularArray(content, 5, zeros_length=0)
    listoffset = regular.toListOffsetArray64(False)
    regular_regular = ak.layout.RegularArray(regular, 3, zeros_length=0)
    listoffset_regular = regular_regular.toListOffsetArray64(False)
    regular_listoffset = ak.layout.RegularArray(listoffset, 3, zeros_length=0)
    listoffset_listoffset = regular_listoffset.toListOffsetArray64(False)
    mask = ak.layout.Index8(np.array([True, False]))
    regular_regular = ak.layout.ByteMaskedArray(mask, regular_regular, valid_when=False)
    listoffset_regular = ak.layout.ByteMaskedArray(
        mask, listoffset_regular, valid_when=False
    )
    regular_listoffset = ak.layout.ByteMaskedArray(
        mask, regular_listoffset, valid_when=False
    )
    listoffset_listoffset = ak.layout.ByteMaskedArray(
        mask, listoffset_listoffset, valid_when=False
    )

    assert (
        str(ak.type(ak.Array(listoffset_listoffset))) == "2 * option[var * var * int64]"
    )
    axis1 = ak.sum(listoffset_listoffset, axis=-1, keepdims=True)
    axis2 = ak.sum(listoffset_listoffset, axis=-2, keepdims=True)
    axis3 = ak.sum(listoffset_listoffset, axis=-3, keepdims=True)
    assert str(ak.type(ak.Array(axis1))) == "2 * option[var * var * int64]"
    assert str(ak.type(ak.Array(axis2))) == "2 * option[var * var * int64]"
    assert str(ak.type(ak.Array(axis3))) == "1 * var * var * int64"

    assert str(ak.type(ak.Array(listoffset_regular))) == "2 * option[var * 5 * int64]"
    axis1 = ak.sum(listoffset_regular, axis=-1, keepdims=True)
    axis2 = ak.sum(listoffset_regular, axis=-2, keepdims=True)
    axis3 = ak.sum(listoffset_regular, axis=-3, keepdims=True)
    assert str(ak.type(ak.Array(axis1))) == "2 * option[var * 1 * int64]"
    assert str(ak.type(ak.Array(axis2))) == "2 * option[var * 5 * int64]"
    assert str(ak.type(ak.Array(axis3))) == "1 * var * 5 * int64"

    assert str(ak.type(ak.Array(regular_listoffset))) == "2 * option[3 * var * int64]"
    axis1 = ak.sum(regular_listoffset, axis=-1, keepdims=True)
    axis2 = ak.sum(regular_listoffset, axis=-2, keepdims=True)
    axis3 = ak.sum(regular_listoffset, axis=-3, keepdims=True)
    assert str(ak.type(ak.Array(axis1))) == "2 * option[var * var * int64]"
    assert str(ak.type(ak.Array(axis2))) == "2 * option[var * var * int64]"
    assert str(ak.type(ak.Array(axis3))) == "1 * var * var * int64"

    assert str(ak.type(ak.Array(regular_regular))) == "2 * option[3 * 5 * int64]"
    axis1 = ak.sum(regular_regular, axis=-1, keepdims=True)
    axis2 = ak.sum(regular_regular, axis=-2, keepdims=True)
    axis3 = ak.sum(regular_regular, axis=-3, keepdims=True)
    assert str(ak.type(ak.Array(axis1))) == "2 * option[var * 1 * int64]"
    assert str(ak.type(ak.Array(axis2))) == "2 * option[var * 5 * int64]"
    assert str(ak.type(ak.Array(axis3))) == "1 * var * 5 * int64"



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_integerindex_null():
    a = ak.Array([[0, 1, 2], None, [5, 6], [7]])
    b = ak.Array([[0, 1, 2], [3, 4], [5, 6], [7]])
    c = ak.Array([[1], [1], [0], [0]])
    d = ak.Array([[1], None, [0], [0]])
    e = ak.Array([[1], None, None, [0]])

    assert ak.to_list(a[c]) == [[1], None, [5], [7]]
    assert ak.to_list(a[d]) == [[1], None, [5], [7]]
    assert ak.to_list(a[e]) == [[1], None, None, [7]]
    assert ak.to_list(b[c]) == [[1], [4], [5], [7]]
    assert ak.to_list(b[d]) == [[1], None, [5], [7]]
    assert ak.to_list(b[e]) == [[1], None, None, [7]]


def test_boolindex_null():
    a = ak.Array([[0, 1, 2], None, [5, 6]])
    b = ak.Array([[0, 1, 2], [3, 4], [5, 6]])
    c = ak.Array([[False, True, False], [False, True], [True, False]])
    d = ak.Array([[False, True, False], None, [True, False]])
    e = ak.Array([[False, True, False], None, None])

    assert ak.to_list(a[c]) == [[1], None, [5]]
    assert ak.to_list(a[d]) == [[1], None, [5]]
    assert ak.to_list(a[e]) == [[1], None, None]
    assert ak.to_list(b[c]) == [[1], [4], [5]]
    assert ak.to_list(b[d]) == [[1], None, [5]]
    assert ak.to_list(b[e]) == [[1], None, None]

    b2 = b.mask[[True, False, True]]
    assert ak.to_list(b2[c]) == [[1], None, [5]]
    assert ak.to_list(b2[d]) == [[1], None, [5]]
    assert ak.to_list(b2[e]) == [[1], None, None]


def test_integerindex_null_more():
    f = ak.Array([[0, None, 2], None, [3, 4], []])
    g1 = ak.Array([[1, 2, None], None, [], [None]])
    g2 = ak.Array([[], None, None, []])
    g3 = ak.Array([[], [], [], []])

    assert ak.to_list(f[g1]) == [[None, 2, None], None, [], [None]]
    assert ak.to_list(f[g2]) == [[], None, None, []]
    assert ak.to_list(f[g3]) == [[], None, [], []]

    a = ak.Array([[0, 1, 2, None], None])
    b = ak.Array([[2, 1, None, 3], None])
    assert ak.to_list(a[b]) == [[2, 1, None, None], None]
    b = ak.Array([[2, 1, None, 3], []])
    assert ak.to_list(a[b]) == [[2, 1, None, None], None]
    b = ak.Array([[2, 1, None, 3], [0, 1]])
    assert ak.to_list(a[b]) == [[2, 1, None, None], None]

    a = ak.Array([[[0, 1, 2, None], None], [[3, 4], [5]], None, [[6]]])
    b = ak.Array([[[2, 1, None, 3], [0, 1]], [[0], None], None, [None]])
    c = ak.Array(
        [
            [[False, True, None, False], [False, True]],
            [[True, False], None],
            None,
            [None],
        ]
    )
    assert ak.to_list(a[b]) == [[[2, 1, None, None], None], [[3], None], None, [None]]
    assert ak.to_list(a[c]) == [[[1, None], None], [[4], None], None, [None]]


def test_silly_stuff():
    a = ak.Array([[0, 1, 2], 3])
    b = [[2], [0]]
    with pytest.raises(ValueError):
        a[b]

    a = ak.Array([[0, 1, 2], [3, 4], [5, 6], [7]])
    ### This should not be allowed. I don't know why it was here.
    # b = ak.Array([[0, 2], None])
    b = ak.Array([[0, 2], None, [1], None])
    assert ak.to_list(a[b]) == [[0, 2], None, [6], None]
    b = ak.Array([[0, 2], None, None, None, None, None])
    with pytest.raises(ValueError):
        a[b]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_getitem():
    content0 = ak.from_iter([[1.1, 2.2, 3.3], [], [4.4, 5.5]], highlevel=False)
    content1 = ak.from_iter(["one", "two", "three", "four", "five"], highlevel=False)
    tags = ak.layout.Index8(np.array([1, 1, 0, 0, 1, 0, 1, 1], dtype=np.int8))

    assert np.asarray(ak.layout.UnionArray8_32.regular_index(tags)).tolist() == [
        0,
        1,
        0,
        1,
        2,
        2,
        3,
        4,
    ]
    assert np.asarray(ak.layout.UnionArray8_32.regular_index(tags)).dtype == np.dtype(
        np.int32
    )
    assert np.asarray(ak.layout.UnionArray8_U32.regular_index(tags)).tolist() == [
        0,
        1,
        0,
        1,
        2,
        2,
        3,
        4,
    ]
    assert np.asarray(ak.layout.UnionArray8_U32.regular_index(tags)).dtype == np.dtype(
        np.uint32
    )
    assert np.asarray(ak.layout.UnionArray8_64.regular_index(tags)).tolist() == [
        0,
        1,
        0,
        1,
        2,
        2,
        3,
        4,
    ]
    assert np.asarray(ak.layout.UnionArray8_64.regular_index(tags)).dtype == np.dtype(
        np.int64
    )

    index = ak.layout.Index32(np.array([0, 1, 0, 1, 2, 2, 4, 3], dtype=np.int32))
    array = ak.layout.UnionArray8_32(tags, index, [content0, content1])
    assert np.asarray(array.tags).tolist() == [1, 1, 0, 0, 1, 0, 1, 1]
    assert np.asarray(array.tags).dtype == np.dtype(np.int8)
    assert np.asarray(array.index).tolist() == [0, 1, 0, 1, 2, 2, 4, 3]
    assert np.asarray(array.index).dtype == np.dtype(np.int32)
    assert type(array.contents) is list
    assert [ak.to_list(x) for x in array.contents] == [
        [[1.1, 2.2, 3.3], [], [4.4, 5.5]],
        ["one", "two", "three", "four", "five"],
    ]
    assert array.numcontents == 2
    assert ak.to_list(array.content(0)) == [[1.1, 2.2, 3.3], [], [4.4, 5.5]]
    assert ak.to_list(array.content(1)) == ["one", "two", "three", "four", "five"]
    assert ak.to_list(array.project(0)) == [[1.1, 2.2, 3.3], [], [4.4, 5.5]]
    assert ak.to_list(array.project(1)) == ["one", "two", "three", "five", "four"]
    repr(array)

    assert ak.to_list(array[0]) == "one"
    assert ak.to_list(array[1]) == "two"
    assert ak.to_list(array[2]) == [1.1, 2.2, 3.3]
    assert ak.to_list(array[3]) == []
    assert ak.to_list(array[4]) == "three"
    assert ak.to_list(array[5]) == [4.4, 5.5]
    assert ak.to_list(array[6]) == "five"
    assert ak.to_list(array[7]) == "four"

    assert ak.to_list(array) == [
        "one",
        "two",
        [1.1, 2.2, 3.3],
        [],
        "three",
        [4.4, 5.5],
        "five",
        "four",
    ]
    assert ak.to_list(array[1:-1]) == [
        "two",
        [1.1, 2.2, 3.3],
        [],
        "three",
        [4.4, 5.5],
        "five",
    ]
    assert ak.to_list(array[2:-2]) == [[1.1, 2.2, 3.3], [], "three", [4.4, 5.5]]
    assert ak.to_list(array[::2]) == ["one", [1.1, 2.2, 3.3], "three", "five"]
    assert ak.to_list(array[::2, 1:]) == ["ne", [2.2, 3.3], "hree", "ive"]
    assert ak.to_list(array[:, :-1]) == [
        "on",
        "tw",
        [1.1, 2.2],
        [],
        "thre",
        [4.4],
        "fiv",
        "fou",
    ]

    content2 = ak.from_iter(
        [{"x": 0, "y": []}, {"x": 1, "y": [1.1]}, {"x": 2, "y": [1.1, 2.2]}],
        highlevel=False,
    )
    content3 = ak.from_iter(
        [
            {"x": 0.0, "y": "zero", "z": False},
            {"x": 1.1, "y": "one", "z": True},
            {"x": 2.2, "y": "two", "z": False},
            {"x": 3.3, "y": "three", "z": True},
            {"x": 4.4, "y": "four", "z": False},
        ],
        highlevel=False,
    )
    array2 = ak.layout.UnionArray8_32(tags, index, [content2, content3])
    assert ak.to_list(array2) == [
        {"x": 0.0, "y": "zero", "z": False},
        {"x": 1.1, "y": "one", "z": True},
        {"x": 0, "y": []},
        {"x": 1, "y": [1.1]},
        {"x": 2.2, "y": "two", "z": False},
        {"x": 2, "y": [1.1, 2.2]},
        {"x": 4.4, "y": "four", "z": False},
        {"x": 3.3, "y": "three", "z": True},
    ]
    assert ak.to_list(array2["x"]) == [0.0, 1.1, 0, 1, 2.2, 2, 4.4, 3.3]
    assert ak.to_list(array2["y"]) == [
        "zero",
        "one",
        [],
        [1.1],
        "two",
        [1.1, 2.2],
        "four",
        "three",
    ]
    assert ak.to_list(array2[:, "y", 1:]) == [
        "ero",
        "ne",
        [],
        [],
        "wo",
        [2.2],
        "our",
        "hree",
    ]
    assert ak.to_list(array2["y", :, 1:]) == [
        "ero",
        "ne",
        [],
        [],
        "wo",
        [2.2],
        "our",
        "hree",
    ]
    with pytest.raises(ValueError) as err:
        array2[:, 1:, "y"]
    assert str(err.value).startswith("in NumpyArray, too many dimensions in slice")
    with pytest.raises(ValueError) as err:
        array2["z"]
    assert str(err.value).startswith('key "z" does not exist (not in record)')

    array3 = ak.layout.UnionArray8_32(tags, index, [content3, content2])
    array4 = ak.layout.UnionArray8_32(
        tags, index, [content0, content1, content2, content3]
    )
    assert set(content2.keys()) == set(["x", "y"])
    assert set(content3.keys()) == set(["x", "y", "z"])
    assert set(array2.keys()) == set(["x", "y"])
    assert set(array3.keys()) == set(["x", "y"])
    assert array4.keys() == []


def test_identities():
    content0 = ak.from_iter([[1.1, 2.2, 3.3], [], [4.4, 5.5]], highlevel=False)
    content1 = ak.from_iter(["one", "two", "three", "four", "five"], highlevel=False)
    tags = ak.layout.Index8(np.array([1, 1, 0, 0, 1, 0, 1, 1], dtype=np.int8))
    index = ak.layout.Index32(np.array([0, 1, 0, 1, 2, 2, 4, 3], dtype=np.int32))
    array = ak.layout.UnionArray8_32(tags, index, [content0, content1])

    array.setidentities()
    assert np.asarray(array.identities).tolist() == [
        [0],
        [1],
        [2],
        [3],
        [4],
        [5],
        [6],
        [7],
    ]
    assert np.asarray(array.content(0).identities).tolist() == [[2], [3], [5]]
    assert np.asarray(array.content(1).identities).tolist() == [[0], [1], [4], [7], [6]]


def test_fromiter():
    builder = ak.layout.ArrayBuilder()

    builder.integer(0)
    builder.integer(1)
    builder.integer(2)
    builder.beginlist()
    builder.endlist()
    builder.beginlist()
    builder.real(1.1)
    builder.endlist()
    builder.beginlist()
    builder.real(1.1)
    builder.real(2.2)
    builder.endlist()
    builder.beginlist()
    builder.real(1.1)
    builder.real(2.2)
    builder.real(3.3)
    builder.endlist()

    assert ak.to_list(builder.snapshot()) == [
        0,
        1,
        2,
        [],
        [1.1],
        [1.1, 2.2],
        [1.1, 2.2, 3.3],
    ]

    assert ak.to_list(
        ak.from_iter([0, 1, 2, [], [1.1], [1.1, 2.2], [1.1, 2.2, 3.3]])
    ) == [0, 1, 2, [], [1.1], [1.1, 2.2], [1.1, 2.2, 3.3]]
    assert ak.to_list(
        ak.from_iter(
            [
                0,
                1,
                2,
                [],
                "zero",
                [1.1],
                "one",
                [1.1, 2.2],
                "two",
                [1.1, 2.2, 3.3],
                "three",
            ]
        )
    ) == [
        0,
        1,
        2,
        [],
        "zero",
        [1.1],
        "one",
        [1.1, 2.2],
        "two",
        [1.1, 2.2, 3.3],
        "three",
    ]
    assert ak.to_list(
        ak.from_json(
            '[0, 1, 2, [], "zero", [1.1], "one", [1.1, 2.2], "two", [1.1, 2.2, 3.3], "three"]'
        )
    ) == [
        0,
        1,
        2,
        [],
        "zero",
        [1.1],
        "one",
        [1.1, 2.2],
        "two",
        [1.1, 2.2, 3.3],
        "three",
    ]
    assert (
        ak.to_json(
            ak.from_json(
                '[0,1,2,[],"zero",[1.1],"one",[1.1,2.2],"two",[1.1,2.2,3.3],"three"]'
            )
        )
        == '[0,1,2,[],"zero",[1.1],"one",[1.1,2.2],"two",[1.1,2.2,3.3],"three"]'
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401

from awkward._v2.tmp_for_testing import v1_to_v2

pytestmark = pytest.mark.skipif(
    ak._util.py27, reason="No Python 2.7 support in Awkward 2.x"
)

primes = [x for x in range(2, 1000) if all(x % n != 0 for n in range(2, x))]


def test():
    data = ak.Array([[1, 3, 5, 4, 2], [], [2, 3, 1], [5]])
    data = v1_to_v2(data.layout)
    assert ak.to_list(data.min(axis=1, initial=4)) == [1, None, 1, 4]
    assert (
        data.typetracer.min(axis=1, initial=4).form == data.min(axis=1, initial=4).form
    )
    assert ak.to_list(data.min(axis=1)) == [1, None, 1, 5]
    assert data.typetracer.min(axis=1).form == data.min(axis=1).form
    assert ak.to_list(data.max(axis=1, initial=4)) == [5, None, 4, 5]
    assert (
        data.typetracer.max(axis=1, initial=4).form == data.max(axis=1, initial=4).form
    )
    assert ak.to_list(data.max(axis=1)) == [5, None, 3, 5]
    assert data.typetracer.max(axis=1).form == data.max(axis=1).form

    data = ak.Array([[1.1, 3.3, 5.5, 4.4, 2.2], [], [2.2, 3.3, 1.1], [5.5]])
    data = v1_to_v2(data.layout)
    assert ak.to_list(data.min(axis=1, initial=4)) == [1.1, None, 1.1, 4]
    assert (
        data.typetracer.min(axis=1, initial=4).form == data.min(axis=1, initial=4).form
    )
    assert ak.to_list(data.min(axis=1)) == [1.1, None, 1.1, 5.5]
    assert data.typetracer.min(axis=1).form == data.min(axis=1).form
    assert ak.to_list(data.max(axis=1, initial=4)) == [5.5, None, 4, 5.5]
    assert (
        data.typetracer.max(axis=1, initial=4).form == data.max(axis=1, initial=4).form
    )
    assert ak.to_list(data.max(axis=1)) == [5.5, None, 3.3, 5.5]
    assert data.typetracer.max(axis=1).form == data.max(axis=1).form

    array = ak.layout.NumpyArray(np.array(primes[2 : 2 * 3 * 5], dtype=np.int32))
    array = data = v1_to_v2(array)
    assert ak.to_list(data.min(axis=-1, initial=4)) == 4
    assert ak.to_list(data.min(axis=-1)) == 5
    assert ak.to_list(data.max(axis=-1, initial=114)) == 114
    assert ak.to_list(data.max(axis=-1)) == 113

    array = ak.layout.NumpyArray(np.array(primes[2 : 2 * 3 * 5], dtype=np.uint64))
    array = data = v1_to_v2(array)
    assert ak.to_list(data.min(axis=-1, initial=4)) == 4
    assert ak.to_list(data.min(axis=-1)) == 5
    assert ak.to_list(data.max(axis=-1, initial=114)) == 114
    assert ak.to_list(data.max(axis=-1)) == 113

    array = ak.layout.NumpyArray(np.array(primes[2 : 2 * 3 * 5], dtype=np.uint32))
    array = data = v1_to_v2(array)
    assert ak.to_list(data.min(axis=-1, initial=4)) == 4
    assert ak.to_list(data.min(axis=-1)) == 5
    assert ak.to_list(data.max(axis=-1, initial=114)) == 114
    assert ak.to_list(data.max(axis=-1)) == 113


@pytest.mark.skip(reason="FIXME: handle date-time case for min/max (@ianna)")
def test_date_time():
    numpy_array = np.array(
        ["2020-07-27T10:41:11", "2019-01-01", "2020-01-01"], "datetime64[s]"
    )

    array = ak.Array(numpy_array)
    assert str(array.type) == "3 * datetime64"
    assert array.tolist() == [
        np.datetime64("2020-07-27T10:41:11"),
        np.datetime64("2019-01-01T00:00:00"),
        np.datetime64("2020-01-01T00:00:00"),
    ]

    array = v1_to_v2(array.layout)
    assert ak.to_list(array) == [
        np.datetime64("2020-07-27T10:41:11"),
        np.datetime64("2019-01-01T00:00:00"),
        np.datetime64("2020-01-01T00:00:00"),
    ]

    assert array.max() == numpy_array[0]
    assert array.min() == numpy_array[1]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable

import json

import awkward as ak
from awkward._v2.types.type import Type
from awkward._v2.forms.form import _parameters_equal


class RecordType(Type):
    def __init__(self, contents, keys, parameters=None, typestr=None):
        if not isinstance(contents, Iterable):
            raise TypeError(
                "{0} 'contents' must be iterable, not {1}".format(
                    type(self).__name__, repr(contents)
                )
            )
        if not isinstance(contents, list):
            contents = list(contents)
        for content in contents:
            if not isinstance(content, Type):
                raise TypeError(
                    "{0} all 'contents' must be Type subclasses, not {1}".format(
                        type(self).__name__, repr(content)
                    )
                )
        if keys is not None and not isinstance(keys, Iterable):
            raise TypeError(
                "{0} 'keys' must be iterable, not {1}".format(
                    type(self).__name__, repr(contents)
                )
            )
        if parameters is not None and not isinstance(parameters, dict):
            raise TypeError(
                "{0} 'parameters' must be of type dict or None, not {1}".format(
                    type(self).__name__, repr(parameters)
                )
            )
        if typestr is not None and not ak._util.isstr(typestr):
            raise TypeError(
                "{0} 'typestr' must be of type string or None, not {1}".format(
                    type(self).__name__, repr(typestr)
                )
            )
        self._contents = contents
        self._keys = keys
        self._parameters = parameters
        self._typestr = typestr

    @property
    def contents(self):
        return self._contents

    @property
    def keys(self):
        return self._keys

    @property
    def is_tuple(self):
        return self._keys is None

    _str_parameters_exclude = ("__categorical__", "__record__")

    def __str__(self):
        if self._typestr is not None:
            out = self._typestr

        else:
            children = [str(x) for x in self._contents]
            params = self._str_parameters()
            name = self.parameter("__record__")

            if params is None:
                if self.is_tuple:
                    if name is None:
                        out = "(" + ", ".join(children) + ")"
                    else:
                        out = name + "[" + ", ".join(children) + "]"
                else:
                    pairs = [k + ": " + v for k, v in zip(self._keys, children)]
                    if name is None:
                        out = "{" + ", ".join(pairs) + "}"
                    else:
                        out = name + "[" + ", ".join(pairs) + "]"

            else:
                if self.is_tuple:
                    if name is None:
                        out = "tuple[[{0}], {1}]".format(", ".join(children), params)
                    else:
                        out = "{0}[{1}, {2}]".format(name, ", ".join(children), params)
                else:
                    if name is None:
                        keys = [json.dumps(x) for x in self._keys]
                        out = "struct[[{0}], [{1}], {2}]".format(
                            ", ".join(keys), ", ".join(children), params
                        )
                    else:
                        pairs = [k + ": " + v for k, v in zip(self._keys, children)]
                        out = "{0}[{1}, {2}]".format(name, ", ".join(pairs), params)

        return self._str_categorical_begin() + out + self._str_categorical_end()

    def __repr__(self):
        args = [repr(self._contents), repr(self._keys)] + self._repr_args()
        return "{0}({1})".format(type(self).__name__, ", ".join(args))

    def __eq__(self, other):
        if isinstance(other, RecordType):
            if self._typestr != other._typestr or not _parameters_equal(
                self._parameters, other._parameters
            ):
                return False

            if self._keys is None and other._keys is None:
                return self._contents == other._contents

            elif self._keys is not None and other._keys is not None:
                if set(self._keys) != set(other._keys):
                    return False
                for key in self._keys:
                    if self.content(key) != other.content(key):
                        return False
                else:
                    return True

            else:
                return False

        else:
            return False

    def index_to_key(self, index):
        return ak._v2.forms.recordform.RecordForm.index_to_key(self, index)

    def key_to_index(self, key):
        return ak._v2.forms.recordform.RecordForm.key_to_index(self, key)

    def haskey(self, key):
        return ak._v2.forms.recordform.RecordForm.haskey(self, key)

    def content(self, index_or_key):
        return ak._v2.forms.recordform.RecordForm.content(self, index_or_key)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401

awkward0 = pytest.importorskip("awkward0")


def test_toawkward0():
    array = ak.from_iter([1.1, 2.2, 3.3, 4.4], highlevel=False)
    assert isinstance(ak.to_awkward0(array), np.ndarray)
    assert ak.to_awkward0(array).tolist() == [1.1, 2.2, 3.3, 4.4]

    array = ak.from_numpy(
        np.arange(2 * 3 * 5).reshape(2, 3, 5), highlevel=False
    ).toRegularArray()
    assert isinstance(ak.to_awkward0(array), awkward0.JaggedArray)
    assert ak.to_awkward0(array).tolist() == [
        [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9], [10, 11, 12, 13, 14]],
        [[15, 16, 17, 18, 19], [20, 21, 22, 23, 24], [25, 26, 27, 28, 29]],
    ]

    array = ak.from_iter([[1.1, 2.2, 3.3], [], [4.4, 5.5]], highlevel=False)
    assert isinstance(ak.to_awkward0(array), awkward0.JaggedArray)
    assert ak.to_awkward0(array).tolist() == [[1.1, 2.2, 3.3], [], [4.4, 5.5]]

    array = ak.layout.ListArray64(
        ak.layout.Index64(np.array([4, 999, 1], dtype=np.int64)),
        ak.layout.Index64(np.array([7, 999, 3], dtype=np.int64)),
        ak.layout.NumpyArray(np.array([3.14, 4.4, 5.5, 123, 1.1, 2.2, 3.3, 321])),
    )
    assert isinstance(ak.to_awkward0(array), awkward0.JaggedArray)
    assert ak.to_awkward0(array).tolist() == [[1.1, 2.2, 3.3], [], [4.4, 5.5]]

    array = ak.from_iter(
        [
            {"x": 0, "y": []},
            {"x": 1.1, "y": [1]},
            {"x": 2.2, "y": [2, 2]},
            {"x": 3.3, "y": [3, 3, 3]},
        ],
        highlevel=False,
    )
    assert isinstance(ak.to_awkward0(array[2]), dict)
    assert ak.to_awkward0(array[2])["x"] == 2.2
    assert isinstance(ak.to_awkward0(array[2])["y"], np.ndarray)
    assert ak.to_awkward0(array[2])["y"].tolist() == [2, 2]

    assert isinstance(ak.to_awkward0(array), awkward0.Table)
    assert ak.to_awkward0(array).tolist() == [
        {"x": 0, "y": []},
        {"x": 1.1, "y": [1]},
        {"x": 2.2, "y": [2, 2]},
        {"x": 3.3, "y": [3, 3, 3]},
    ]

    array = ak.from_iter(
        [(0, []), (1.1, [1]), (2.2, [2, 2]), (3.3, [3, 3, 3])], highlevel=False
    )
    assert isinstance(ak.to_awkward0(array), awkward0.Table)
    assert ak.to_awkward0(array).tolist() == [
        (0, []),
        (1.1, [1]),
        (2.2, [2, 2]),
        (3.3, [3, 3, 3]),
    ]
    assert isinstance(ak.to_awkward0(array[2]), tuple)
    assert ak.to_awkward0(array[2])[0] == 2.2
    assert ak.to_awkward0(array[2])[1].tolist() == [2, 2]

    array = ak.from_iter(
        [0.0, [], 1.1, [1], 2.2, [2, 2], 3.3, [3, 3, 3]], highlevel=False
    )
    assert isinstance(ak.to_awkward0(array), awkward0.UnionArray)
    assert ak.to_awkward0(array).tolist() == [
        0.0,
        [],
        1.1,
        [1],
        2.2,
        [2, 2],
        3.3,
        [3, 3, 3],
    ]

    array = ak.from_iter([1.1, 2.2, None, None, 3.3, None, 4.4], highlevel=False)
    assert isinstance(ak.to_awkward0(array), awkward0.IndexedMaskedArray)
    assert ak.to_awkward0(array).tolist() == [1.1, 2.2, None, None, 3.3, None, 4.4]

    content = ak.layout.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    index = ak.layout.Index64(np.array([3, 2, 2, 5, 0], dtype=np.int64))
    array = ak.layout.IndexedArray64(index, content)
    assert isinstance(ak.to_awkward0(array), awkward0.IndexedArray)
    assert ak.to_awkward0(array).tolist() == [3.3, 2.2, 2.2, 5.5, 0.0]


def test_fromawkward0():
    array = np.array([1.1, 2.2, 3.3, 4.4, 5.5])
    assert isinstance(ak.from_awkward0(array), ak.highlevel.Array)
    assert isinstance(ak.from_awkward0(array, highlevel=False), ak.layout.NumpyArray)
    assert ak.to_list(array) == [1.1, 2.2, 3.3, 4.4, 5.5]

    array = (123, np.array([1.1, 2.2, 3.3]))
    assert isinstance(ak.from_awkward0(array), ak.highlevel.Record)
    assert isinstance(ak.from_awkward0(array).layout, ak.layout.Record)
    assert ak.to_list(ak.from_awkward0(array)) == (123, [1.1, 2.2, 3.3])

    array = {"x": 123, "y": np.array([1.1, 2.2, 3.3])}
    assert isinstance(ak.from_awkward0(array), ak.highlevel.Record)
    assert isinstance(ak.from_awkward0(array).layout, ak.layout.Record)
    assert ak.to_list(ak.from_awkward0(array)) == {"x": 123, "y": [1.1, 2.2, 3.3]}

    array = awkward0.fromiter([[1.1, 2.2, 3.3], [], [4.4, 5.5]])
    assert isinstance(
        ak.from_awkward0(array, highlevel=False),
        (
            ak.layout.ListOffsetArray32,
            ak.layout.ListOffsetArrayU32,
            ak.layout.ListOffsetArray64,
        ),
    )
    assert ak.to_list(ak.from_awkward0(array)) == [[1.1, 2.2, 3.3], [], [4.4, 5.5]]

    array = awkward0.fromiter(
        [{"x": 0, "y": []}, {"x": 1.1, "y": [1]}, {"x": 2.2, "y": [2, 2]}]
    )
    assert isinstance(ak.from_awkward0(array, highlevel=False), ak.layout.RecordArray)
    assert not ak.from_awkward0(array, highlevel=False).istuple
    assert ak.from_awkward0(array).layout.keys() == ["x", "y"]
    assert ak.to_list(ak.from_awkward0(array)) == [
        {"x": 0, "y": []},
        {"x": 1.1, "y": [1]},
        {"x": 2.2, "y": [2, 2]},
    ]

    array = awkward0.Table([0.0, 1.1, 2.2], awkward0.fromiter([[], [1], [2, 2]]))
    assert isinstance(ak.from_awkward0(array, highlevel=False), ak.layout.RecordArray)
    assert ak.from_awkward0(array, highlevel=False).istuple
    assert ak.from_awkward0(array).layout.keys() == ["0", "1"]
    assert ak.to_list(ak.from_awkward0(array)) == [(0.0, []), (1.1, [1]), (2.2, [2, 2])]

    array = awkward0.fromiter([0.0, [], 1.1, [1], 2.2, [2, 2], 3.3, [3, 3, 3]])
    assert isinstance(
        ak.from_awkward0(array, highlevel=False),
        (ak.layout.UnionArray8_32, ak.layout.UnionArray8_U32, ak.layout.UnionArray8_64),
    )
    assert ak.to_list(ak.from_awkward0(array)) == [
        0.0,
        [],
        1.1,
        [1],
        2.2,
        [2, 2],
        3.3,
        [3, 3, 3],
    ]

    array = awkward0.fromiter([1.1, 2.2, None, None, 3.3, None, 4.4])
    assert isinstance(
        ak.from_awkward0(array, highlevel=False), ak.layout.ByteMaskedArray
    )
    assert ak.to_list(ak.from_awkward0(array)) == [1.1, 2.2, None, None, 3.3, None, 4.4]

    array = awkward0.fromiter(["hello", "you", "guys"])
    assert isinstance(
        ak.from_awkward0(array, highlevel=False),
        (
            ak.layout.ListArray32,
            ak.layout.ListArrayU32,
            ak.layout.ListArray64,
            ak.layout.ListOffsetArray32,
            ak.layout.ListOffsetArrayU32,
            ak.layout.ListOffsetArray64,
        ),
    )
    assert ak.from_awkward0(array, highlevel=False).parameters["__array__"] in (
        "string",
        "bytestring",
    )
    assert ak.from_awkward0(array, highlevel=False).content.parameters["__array__"] in (
        "char",
        "byte",
    )
    assert ak.to_list(ak.from_awkward0(array)) == ["hello", "you", "guys"]

    class Point(object):
        def __init__(self, x, y):
            self.x, self.y = x, y

        def __repr__(self):
            return "Point({0}, {1})".format(self.x, self.y)

    array = awkward0.fromiter([Point(1.1, 10), Point(2.2, 20), Point(3.3, 30)])
    assert ak.to_list(ak.from_awkward0(array)) == [
        {"x": 1.1, "y": 10},
        {"x": 2.2, "y": 20},
        {"x": 3.3, "y": 30},
    ]
    assert "__record__" in ak.from_awkward0(array).layout.parameters

    array = awkward0.ChunkedArray(
        [
            awkward0.fromiter([[1.1, 2.2, 3.3], [], [4.4, 5.5]]),
            awkward0.fromiter([[6.6]]),
            awkward0.fromiter([[7.7, 8.8], [9.9, 10.0, 11.1, 12.2]]),
        ]
    )
    assert ak.to_list(ak.from_awkward0(array)) == [
        [1.1, 2.2, 3.3],
        [],
        [4.4, 5.5],
        [6.6],
        [7.7, 8.8],
        [9.9, 10.0, 11.1, 12.2],
    ]

    def generate1():
        return awkward0.fromiter([[1.1, 2.2, 3.3], [], [4.4, 5.5]])

    def generate2():
        return awkward0.fromiter([[6.6]])

    def generate3():
        return awkward0.fromiter([[7.7, 8.8], [9.9, 10.0, 11.1, 12.2]])

    array = awkward0.ChunkedArray(
        [
            awkward0.VirtualArray(generate1),
            awkward0.VirtualArray(generate2),
            awkward0.VirtualArray(generate3),
        ]
    )
    assert ak.to_list(ak.from_awkward0(array)) == [
        [1.1, 2.2, 3.3],
        [],
        [4.4, 5.5],
        [6.6],
        [7.7, 8.8],
        [9.9, 10.0, 11.1, 12.2],
    ]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import awkward as ak
from awkward._v2.contents.content import NestedIndexError
from awkward._v2.forms.form import Form, _parameters_equal, nonvirtual


class EmptyForm(Form):
    def __init__(self, has_identifier=False, parameters=None, form_key=None):
        self._init(has_identifier, parameters, form_key)

    def __repr__(self):
        args = self._repr_args()
        return "{0}({1})".format(type(self).__name__, ", ".join(args))

    def _tolist_part(self, verbose, toplevel):
        return self._tolist_extra({"class": "EmptyArray"}, verbose)

    def _type(self, typestrs):
        return ak._v2.types.unknowntype.UnknownType(
            self._parameters,
            ak._util.gettypestr(self._parameters, typestrs),
        )

    def __eq__(self, other):
        return (
            isinstance(other, EmptyForm)
            and self._has_identifier == other._has_identifier
            and self._form_key == other._form_key
        )

    def generated_compatibility(self, other):
        other = nonvirtual(other)

        if other is None:
            return True

        elif isinstance(other, EmptyForm):
            return _parameters_equal(self._parameters, other._parameters)

        else:
            return False

    def _getitem_range(self):
        return EmptyForm(
            has_identifier=self._has_identifier,
            parameters=self._parameters,
            form_key=None,
        )

    def _getitem_field(self, where, only_fields=()):
        raise NestedIndexError(self, where, "not an array of records")

    def _getitem_fields(self, where, only_fields=()):
        raise NestedIndexError(self, where, "not an array of records")

    def _carry(self, allow_lazy):
        return EmptyForm(
            has_identifier=self._has_identifier,
            parameters=self._parameters,
            form_key=None,
        )

    @property
    def purelist_isregular(self):
        return True

    @property
    def purelist_depth(self):
        return 1

    @property
    def minmax_depth(self):
        return (1, 1)

    @property
    def branch_depth(self):
        return (False, 1)

    @property
    def keys(self):
        return []

    @property
    def dimension_optiontype(self):
        return False



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import sys

import awkward as ak

from awkward._typeparser.generated_parser import Lark_StandAlone, Transformer


class TreeToJson(Transformer):
    def string(self, s):
        (s,) = s
        if sys.version_info[0] == 2:
            s = s.encode("utf-8")
        return s[1:-1]

    def number(self, n):
        (n,) = n
        if "." in n:
            return float(n)
        else:
            return int(n)

    list_obj = list
    pair = tuple
    dict_obj = dict

    def null(self, s):
        return None

    def true(self, s):
        return True

    def false(self, s):
        return False


def toast(ptnode, high_level, categorical):
    if ptnode.__class__.__name__ == "Token":
        return ptnode.value
    elif ptnode.data == "start":
        return toast(ptnode.children[0], high_level, categorical)
    elif ptnode.data == "input":
        assert len(ptnode.children) == 1
        return toast(ptnode.children[0], high_level, categorical)
    elif ptnode.data == "predefined_typestr":
        if ptnode.children[0] == "string":
            parms = {"__array__": "string"}
            if categorical:
                parms.update({"__categorical__": True})
                categorical = False
            return ak.types.ListType(
                ak.types.PrimitiveType(
                    "uint8", parameters={"__array__": "char"}, typestr="char"
                ),
                parameters=parms,
                typestr="string",
            )
        elif ptnode.children[0] == "char":
            parms = {"__array__": "char"}
            if categorical:
                parms.update({"__categorical__": True})
                categorical = False
            return ak.types.PrimitiveType("uint8", parameters=parms, typestr="char")
        elif ptnode.children[0] == "byte":
            parms = {"__array__": "byte"}
            if categorical:
                parms.update({"__categorical__": True})
                categorical = False
            return ak.types.PrimitiveType("uint8", parameters=parms, typestr="byte")
        elif ptnode.children[0] == "bytes":
            parms = {"__array__": "bytestring"}
            if categorical:
                parms.update({"__categorical__": True})
                categorical = False
            return ak.types.ListType(
                ak.types.PrimitiveType(
                    "uint8", parameters={"__array__": "byte"}, typestr="byte"
                ),
                parameters=parms,
                typestr="bytes",
            )
        else:
            raise Exception("Unhandled typestring {0}".format(ptnode.children[0]))
    elif ptnode.data == "primitive":
        if len(ptnode.children) == 1:
            parms = {}
            if categorical:
                parms.update({"__categorical__": True})
                categorical = False
            return ak.types.PrimitiveType(
                toast(ptnode.children[0], high_level, False), parameters=parms
            )
        elif len(ptnode.children) == 2:
            parms = toast(ptnode.children[1], high_level, False)
            if categorical:
                parms.update({"__categorical__": True})
                categorical = False
            return ak.types.PrimitiveType(
                toast(ptnode.children[0], high_level, categorical),
                parms,
            )
        else:
            raise Exception("Unhandled PrimitiveType node")
    elif ptnode.data == "categories":
        assert high_level is True
        return toast(ptnode.children[0], high_level, True)
    elif ptnode.data == "unknown":
        if len(ptnode.children) == 0:
            parms = {}
            if categorical:
                parms.update({"__categorical__": True})
                categorical = False
            return ak.types.UnknownType(parameters=parms)
        elif len(ptnode.children) == 1:
            parms = toast(ptnode.children[0], high_level, False)
            if categorical:
                parms.update({"__categorical__": True})
                categorical = False
            return ak.types.UnknownType(parameters=parms)
        else:
            raise Exception("Unhandled UnknownType node")
    elif ptnode.data == "listtype":
        return toast(ptnode.children[0], high_level, categorical)
    elif ptnode.data == "list_single":
        parms = {}
        if categorical:
            parms.update({"__categorical__": True})
            categorical = False
        return ak.types.ListType(
            toast(ptnode.children[0], high_level, False), parameters=parms
        )
    elif ptnode.data == "list_parm":
        parms = toast(ptnode.children[1], high_level, False)
        if categorical:
            parms.update({"__categorical__": True})
            categorical = False
        return ak.types.ListType(
            toast(ptnode.children[0], high_level, categorical), parms
        )
    elif ptnode.data == "uniontype":
        return toast(ptnode.children[0], high_level, categorical)
    elif ptnode.data == "union_single":
        parms = {}
        if categorical:
            parms.update({"__categorical__": True})
            categorical = False
        content_list = []
        for node in ptnode.children:
            content_list.append(toast(node, high_level, False))
        return ak.types.UnionType(content_list, parameters=parms)
    elif ptnode.data == "union_parm":
        parms = toast(ptnode.children[-1], high_level, False)
        if categorical:
            parms.update({"__categorical__": True})
            categorical = False
        content_list = []
        for node in ptnode.children[:-1]:
            content_list.append(toast(node, high_level, False))
        return ak.types.UnionType(content_list, parms)
    elif ptnode.data == "optiontype":
        return toast(ptnode.children[0], high_level, categorical)
    elif ptnode.data == "option_single":
        parms = {}
        if categorical:
            parms.update({"__categorical__": True})
            categorical = False
        return ak.types.OptionType(
            toast(ptnode.children[0], high_level, False), parameters=parms
        )
    elif ptnode.data == "option_parm":
        parms = toast(ptnode.children[1], high_level, False)
        if categorical:
            parms.update({"__categorical__": True})
            categorical = False
        return ak.types.OptionType(
            toast(ptnode.children[0], high_level, False),
            parameters=parms,
        )
    elif ptnode.data == "option_highlevel":
        assert high_level
        parms = {}
        if categorical:
            parms.update({"__categorical__": True})
            categorical = False
        return ak.types.OptionType(
            toast(ptnode.children[0], high_level, False), parameters=parms
        )
    elif ptnode.data == "record":
        return toast(ptnode.children[0], high_level, categorical)
    elif ptnode.data == "record_tuple":
        parms = {}
        if categorical:
            parms.update({"__categorical__": True})
            categorical = False
        content_list = []
        for node in ptnode.children:
            content_list.append(toast(node, high_level, categorical))
        return ak.types.RecordType(tuple(content_list), parameters=parms)
    elif ptnode.data == "record_dict":
        parms = {}
        if categorical:
            parms.update({"__categorical__": True})
            categorical = False
        content_types = []
        content_keys = []
        for i in range(0, len(ptnode.children), 2):
            content_keys.append(ptnode.children[i])
            content_types.append(toast(ptnode.children[i + 1], high_level, categorical))
        return ak.types.RecordType(content_types, content_keys, parameters=parms)
    elif ptnode.data == "record_tuple_param":
        parms = toast(ptnode.children[-1], high_level, False)
        if categorical:
            parms.update({"__categorical__": True})
            categorical = False
        content_list = []
        for node in ptnode.children[:-1]:
            content_list.append(toast(node, high_level, False))
        return ak.types.RecordType(tuple(content_list), parameters=parms)
    elif ptnode.data == "record_struct":
        parms = toast(ptnode.children[-1], high_level, False)
        if categorical:
            parms.update({"__categorical__": True})
            categorical = False
        content_list = []
        content_keys = []
        for node in ptnode.children[:-1]:
            if isinstance(node, str):
                content_keys.append(node)
            else:
                content_list.append(toast(node, high_level, False))
        return ak.types.RecordType(
            tuple(content_list),
            keys=content_keys,
            parameters=parms,
        )
    elif ptnode.data == "record_highlevel":
        assert high_level
        parms = {"__record__": ptnode.children[0]}
        if categorical:
            parms.update({"__categorical__": True})
            categorical = False
        content_list = []
        content_keys = []
        for node in ptnode.children[1:]:
            if isinstance(node, str):
                content_keys.append(node)
            else:
                content_list.append(toast(node, high_level, False))
        return ak.types.RecordType(
            tuple(content_list),
            keys=content_keys,
            parameters=parms,
        )
    elif ptnode.data == "regular":
        assert (len(ptnode.children)) == 1
        return toast(ptnode.children[0], high_level, categorical)
    elif ptnode.data == "regular_inparm":
        assert len(ptnode.children) == 2
        if high_level:
            return ak.types.ArrayType(
                toast(ptnode.children[1], high_level, categorical), ptnode.children[0]
            )
        return ak.types.RegularType(
            toast(ptnode.children[1], high_level, categorical), ptnode.children[0]
        )
    elif ptnode.data == "regular_outparm":
        assert len(ptnode.children) == 3
        parms = toast(ptnode.children[2], high_level, False)
        if categorical:
            parms.update({"__categorical__": True})
            categorical = False
        return ak.types.RegularType(
            toast(ptnode.children[1], high_level, False),
            ptnode.children[0],
            parms,
        )
    elif ptnode.data == "def_option":
        assert len(ptnode.children) == 1
        return ptnode.children[0]
    elif ptnode.data == "options":
        assert len(ptnode.children) == 1
        return toast(ptnode.children[0], high_level, categorical)
    else:
        raise Exception("Unhandled node")


def from_datashape(typestr, high_level=False):
    parseobj = Lark_StandAlone(transformer=TreeToJson())
    return toast(parseobj.parse(typestr), high_level, False)
